#include<linux/init.h>                  // Basic Initialization
#include<linux/module.h>                // Needed for all the modules.
#include<linux/kernel.h>                // Allow to use the printk function.
#include<linux/device.h>                // The device creation API's are defined here.
#include<linux/cdev.h>                  // The cdev structure is defined in it.
#include<linux/fs.h>                    // To register the device. 
#include<linux/slab.h>                  // Needed as kmalloc() is declared in it.
#include<linux/string.h>                // Needed to use string functions.
#include<linux/uaccess.h>               // Needed by segment descriptor.
#include<linux/semaphore.h>             // The semaphore functions are defined in it. 
#include<linux/delay.h>                 // To use delay function ssleep().

MODULE_LICENSE("GPL");                                  	 	// This is a GPL-licesed module.        
MODULE_AUTHOR("FOSSEE TEAM");                           	 	// Macro used to give author name to a module.
MODULE_DESCRIPTION("Synchronization in the device driver.\n");          // Macro gives short description of a module.

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////// 
#define size 25                  // The memory size is defined here.

static dev_t device_num;         // Global variable that holds the device number.
static struct class *cl;         // structure creating the class for a device.
static char *ptr = NULL;         // The character pointer intialiazed to NULL value. 
  
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* cdev structure for a character device .*/

static struct cdev my_cdev;    

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* semaphore for syncronization of a character device .*/

static struct semaphore lock; 

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The prototypes of the file operations such as open, read, write and release that implemented in this driver */

static int mydevice_open (struct inode *inod, struct file *filp);
static int mydevice_release (struct inode *inod, struct file *filp);
static ssize_t mydevice_read (struct file *filp, char __user *buff, size_t len, loff_t *loff);
static ssize_t mydevice_write (struct file *filp, const char __user *buff, size_t len, loff_t *loff);

  
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The file_operation structure of a device.*/

static struct file_operations fops=
{
.read= mydevice_read,
.write= mydevice_write,
.open= mydevice_open,
.release= mydevice_release,
};


/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* Initialization function */

static int init_function(void)
{
	printk(KERN_INFO "my_driver : Inside the init function of a driver.\n");


/********************************************************************************************************************
int alloc_chrdev_region(dev_t *dev, unsigned int firstminor, unsigned int count, char *name) 

Description : It is use for the dynamic allocation of a device number by the kernel.
*********************************************************************************************************************/

	if(alloc_chrdev_region(&device_num, 0, 1, "my_device") < 0)
	{	
		printk(KERN_INFO "my_driver : The device number allocation fails.\n");
		return -1;
	}

/******************************************************************************************************************** 
The important macros are as follows : 
1. MAJOR(dev_t dev) 		:	Extracts the major number from device number.
2. MINOR(dev_t dev)     	:	Extracts the minor number from device number.
*********************************************************************************************************************/
	
	printk(KERN_INFO "my_driver : The device number allocation done successfully.\n");	
	printk(KERN_INFO "my_driver : Major and minor numbers of my_device are : %d, %d\n", MAJOR(device_num), 		  MINOR(device_num));

/******************************************************************************************************************** 
struct class *class_create(struct module *owner, const char *name)

Description - This function use to create a new class under sysfs file system.
*********************************************************************************************************************/
	
	if((cl = class_create(THIS_MODULE,"my_class")) == NULL)
	{
		
		printk(KERN_INFO "my_driver : The class creation fails here.\n");	
		unregister_chrdev_region(device_num,1);
		return -1;
	}
	else
	{
		printk(KERN_INFO "my_driver : The class is created successfully here.\n");

	}

/******************************************************************************************************************** 
struct device *device_create(struct class *class, struct device *parent, dev_t devt, void *drvdata,  const char *fmt,.)

Description - It creates a device and registers it with sysfs and creates the device node under dev directory.
*********************************************************************************************************************/
	
	if(device_create(cl,NULL,device_num,NULL,"new_device") == NULL)
	{
		printk(KERN_INFO "my_driver : The device node creation fails here.\n");

/******************************************************************************************************************** 
void class_destroy (struct class *class)

Description - It destroys a struct class structure.
*********************************************************************************************************************/		
		class_destroy(cl);
		unregister_chrdev_region(device_num,1);
		return -1;
	}
	else 
        {
		printk(KERN_INFO "my_driver : The device node is created successfully.\n");
	
	}

/******************************************************************************************************************** 
void cdev_init (struct cdev *cdev, const struct file_operations *fops)	
Description — initialize a cdev structure 

int cdev_add (struct cdev *p, dev_t dev, unsigned count)	
Description — add a char device to the system 
*********************************************************************************************************************/		
	cdev_init(&my_cdev,&fops);
	if(cdev_add(&my_cdev,device_num,1) == -1)
	{
		device_destroy(cl,device_num);
		class_destroy(cl);
			   unregister_chrdev_region(device_num,1);
		return -1;
	}
	else
	{
		printk(KERN_INFO "my_driver : The cdev_structure initialization done successfully.\n");
			
	}
/*********************************************************************************************************************
void *kmalloc(size_t size, int flags) 

Description - it is use to dynamically allocate the kernel memory 
*********************************************************************************************************************/
	ptr = (char*)kmalloc(size,GFP_ATOMIC);     
	if(ptr == NULL)
	{
		printk(KERN_INFO "my_driver : The dynamic memory allocation fails.\n");	
		return -ENOMEM;
	}	
	else
	{
		printk(KERN_INFO "my_driver : The dynamic memory allocation done successfully.\n");
		
	}

/*********************************************************************************************************************
sema_init(struct semaphore *, int) 

This function is used to initialize the dynamically created semaphore.
*********************************************************************************************************************/
	
	sema_init(&lock,0);
	printk(KERN_INFO "my_driver : The semaphore initializes successfully.\n");
		
	return(0);

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* the open function of a device */

static int mydevice_open (struct inode *inod, struct file *filp)
{
	printk(KERN_INFO "my_driver : The new_device opened successfully.\n");
	
	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The write function defination. */

static ssize_t mydevice_write (struct file *filp, const char __user *buff, size_t len, loff_t *loff)
{
/******************************************************************************************************************** 
unsigned long copy_from_user (	void *to, const void __user *from, unsigned long n)

Description - copy data from the user space to the kernel space. 
*********************************************************************************************************************/

	if(len > 25)
	{
		printk(KERN_INFO "my_driver : The length is greater than the kernel buffer len.\n");
		return EFAULT; 

	}

	if(copy_from_user(ptr,buff,len))
	{
		return -EFAULT;
	}	
	else
	{
		printk(KERN_INFO "my_driver : The write to the device successfully...\n");
		printk("my_driver : the data write to device(ptr) : %s\n",buff);
		
	}
	ssleep(10);
	printk(KERN_INFO "my_driver : The write function goes in sleep.\n");

/*********************************************************************************************************************
up(struct semaphore *) 

Releases the given semaphore and wakes a waiting task, if any.
*********************************************************************************************************************/
	up(&lock);
	printk(KERN_INFO "my_driver : The write function releases the semaphore.\n");

	return len;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The read function defination */
static ssize_t mydevice_read (struct file *filp, char __user *buff, size_t len, loff_t *loff)
{

/******************************************************************************************************************** 
unsigned long copy_to_user (void __user *to, const void *from, unsigned long n)

Description - copy data from the kernel space to the user space. 
*********************************************************************************************************************/
		
		printk(KERN_INFO "my_driver : Inside the my_device read function..\n");

/*********************************************************************************************************************
down_interruptible(struct semaphore *) 

It is used to acquire the given semaphore with interruptible sleep.
*********************************************************************************************************************/
		
		down_interruptible(&lock);
		printk(KERN_INFO "my_driver : Now the read function got the semaphore.\n");
		if(copy_to_user(buff,ptr,len))
		{
			printk(KERN_INFO "my_driver : The data unable to read.\n");
			return -EFAULT;
		}	
		else
		{
			printk(KERN_INFO "my_driver : The read from the device successfully..\n");
			printk("my_driver : the data read from the device(ptr) : %s\n",buff);
			
		}
		return(len);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The close function of a device */

static int mydevice_release(struct inode *inod, struct file *filp)
{
	printk(KERN_INFO "my_driver : the new_device is closed successfully.\n");
	
	return(0);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/* The cleanup function */

static void exit_function(void)
{

	printk(KERN_INFO "my_driver : Inside the exit function of the driver.\n");

/******************************************************************************************************************** 
void cdev_del(struct cdev *cdev)

Description - This function removes the cdev structure from the system.
*********************************************************************************************************************/		
	
	cdev_del(&my_cdev);
	printk(KERN_INFO "my_driver : The cdev structure removed from the system.\n");
	

/******************************************************************************************************************** 
void device_destroy(struct class *class,dev_t devt).

Description - This function unregisters and cleans up a device that was created with a call to device_create. 
*********************************************************************************************************************/		
	device_destroy(cl,device_num);
	printk(KERN_INFO "my_driver : The device removed from the /dev directory and the system.\n");

	class_destroy(cl);
	printk(KERN_INFO "my_driver : The class of the device removed from the system.\n");
	
/*******************************************************************************************************************
void unregister_chrdev_region(dev_t first,unsigned int count).
Description - It is used to free our device number.
	
********************************************************************************************************************/
	unregister_chrdev_region(device_num, 1);
	printk(KERN_INFO "my_driver : The device is unregistered successfully.\n");

/********************************************************************************************************************
void kfree(const void *ptr) 
Description - it is used to freeing the memory allocated by kmalloc. 
********************************************************************************************************************/
	if(ptr != NULL)
	{
	        kfree(ptr);
		printk(KERN_INFO "my_driver : allocted kernel memory releases successfully.\n");	
	
	}
	ptr = NULL;
}

module_init(init_function);
module_exit(exit_function);



